import sys
from typing import Callable, Mapping, Union

import numpy as np

from .partition import Partition

ParticleDataT = Mapping[str, np.ndarray]


def exchange(
    partition: Partition,
    data: dict,
    key: str,
    local_keys: np.ndarray,
    *,
    verbose: bool = False,
    filter_key: Union[int, Callable[[np.ndarray], np.ndarray]] = None,
    do_all2all: bool = False,
    replace_notfound_key: int = None,
):
    """Distribute data among neighboring ranks and all2all by a key

    This function will assign data to the rank that owns the key. The keys that the
    local rank owns are given by ``local_keys``, which should be unique. The keys of the
    data that the local rank currently has is in ``data[key]``. Certain values can be
    ignored by setting filter_key to that value or by setting filter_key to a
    (vectorized) function that returns ``True`` for keys that should be redistributed
    and ``False`` for keys that should be ignored.

    Parameters
    ----------


    Returns
    -------
    """

    if not do_all2all and partition.comm_neighbor is None:
        raise RuntimeError(
            "Cannot exchange data between neighbors if no neighbor topology was "
            "created. Either create partition with ``create_neighbor_topo=True`` "
            "or set ``do_all2all=True``"
        )
    comm = partition.comm
    rank = partition.rank
    nranks = partition.nranks
    if nranks == 1:
        return data

    if do_all2all:
        # exchange particles with all ranks
        exchange_comm = comm
        exchange_nranks = nranks
        exchange_Alltoall = exchange_comm.Alltoall
        exchange_Alltoallv = exchange_comm.Alltoallv
        exchange_Allgather = exchange_comm.Allgather
        exchange_Allgatherv = exchange_comm.Allgatherv

    else:
        # exchange particles with the neighboring ranks
        exchange_comm = partition.comm_neighbor
        exchange_nranks = len(partition.neighbor_ranks)
        exchange_Alltoall = exchange_comm.Neighbor_alltoall
        exchange_Alltoallv = exchange_comm.Neighbor_alltoallv
        exchange_Allgather = exchange_comm.Neighbor_allgather
        exchange_Allgatherv = exchange_comm.Neighbor_allgatherv

    localcount = len(data[key])
    data_keys = np.unique(data[key])
    if filter_key is not None:
        if callable(filter_key):
            data_keys = data_keys[filter_key(data_keys)]
        else:
            data_keys = data_keys[data_keys != filter_key]

    # find local matches
    islocal = np.isin(data_keys, local_keys, assume_unique=True)
    nonlocal_data = data_keys[~islocal]

    # communicate nonlocal descendants
    local_orphan_count = np.array([len(nonlocal_data)], dtype=np.int32)
    orphan_counts = np.empty(exchange_nranks, dtype=np.int32)
    exchange_Allgather(local_orphan_count, orphan_counts)
    total_orphan_count = np.sum(orphan_counts)
    orphan_offsets = np.insert(np.cumsum(orphan_counts)[:-1], 0, 0)
    orphan_data = np.empty(total_orphan_count, dtype=nonlocal_data.dtype)
    orphan_ranks = np.empty(total_orphan_count, dtype=np.int32)
    for i in range(exchange_nranks):
        low = orphan_offsets[i]
        high = low + orphan_counts[i]
        orphan_ranks[low:high] = i
    exchange_Allgatherv(
        nonlocal_data,
        [orphan_data, (orphan_counts, orphan_offsets), nonlocal_data.dtype.char],
    )

    if verbose > 1:
        for i in range(nranks):
            if rank == i:
                print(f"Debug Desc Exchange (nonlocal), rank {i}")
                print(f" - send {local_orphan_count}")
                print(f" - recv {orphan_counts}")
                print(f"", flush=True)
            comm.Barrier()

    # check if we have any of them
    orphan_islocal = np.isin(orphan_data, local_keys)

    # ask
    orphan_requests_send = orphan_data[orphan_islocal]
    orphan_requests_send_ranks = orphan_ranks[orphan_islocal]
    orphan_requests_send_offsets = np.searchsorted(
        orphan_requests_send_ranks, np.arange(exchange_nranks)
    )
    orphan_requests_send_counts = (
        np.append(orphan_requests_send_offsets[1:], len(orphan_requests_send))
        - orphan_requests_send_offsets
    )
    orphan_requests_recv_counts = np.empty_like(orphan_requests_send_counts)
    exchange_Alltoall(orphan_requests_send_counts, orphan_requests_recv_counts)
    orphan_requests_recv_total = np.sum(orphan_requests_recv_counts)
    orphan_requests_recv_offsets = np.insert(
        np.cumsum(orphan_requests_recv_counts)[:-1], 0, 0
    )
    orphan_requests_recv = np.empty(
        orphan_requests_recv_total, dtype=orphan_requests_send.dtype
    )
    exchange_Alltoallv(
        [
            orphan_requests_send,
            (orphan_requests_send_counts, orphan_requests_send_offsets),
            orphan_requests_send.dtype.char,
        ],
        [
            orphan_requests_recv,
            (orphan_requests_recv_counts, orphan_requests_recv_offsets),
            orphan_requests_recv.dtype.char,
        ],
    )

    if verbose > 1:
        for i in range(nranks):
            if rank == i:
                print(f"Debug Desc Exchange (request), rank {i}")
                print(f" - will request {np.sum(orphan_islocal)} in total")
                print(f" - send req {orphan_requests_send_counts}")
                print(f" - recv req {orphan_requests_recv_counts}")
                print(f"", flush=True)
            comm.Barrier()

    # verify that we don't aks ourselves for particles
    if do_all2all and (
        orphan_requests_send_counts[rank] != 0 or orphan_requests_recv_counts[rank] != 0
    ):
        print(
            f"Error in exchange: rank {rank} is asking itself for an orphan halo: "
            f"{orphan_requests_send_counts[rank]}/{orphan_requests_recv_counts[rank]}",
            file=sys.stderr,
            flush=True,
        )
        comm.Abort()

    # prepare data to send
    orphan_requests_indices = []
    orphan_requests_mask = np.zeros(localcount, dtype=np.bool_)
    for i in range(exchange_nranks):
        req = orphan_requests_recv[
            orphan_requests_recv_offsets[i] : orphan_requests_recv_offsets[i]
            + orphan_requests_recv_counts[i]
        ]
        mask = np.isin(data[key], req)
        orphan_requests_indices.append(np.nonzero(mask)[0])
        orphan_requests_mask |= mask
    orphan_requests_send_counts = np.array(
        [len(i) for i in orphan_requests_indices], dtype=np.int32
    )
    orphan_requests_recv_counts = np.empty_like(orphan_requests_send_counts)
    exchange_Alltoall(orphan_requests_send_counts, orphan_requests_recv_counts)
    orphan_requests_recv_total = np.sum(orphan_requests_recv_counts)
    orphan_requests_send_offsets = np.insert(
        np.cumsum(orphan_requests_send_counts)[:-1], 0, 0
    )
    orphan_requests_recv_offsets = np.insert(
        np.cumsum(orphan_requests_recv_counts)[:-1], 0, 0
    )
    orphan_requests_indices = np.concatenate(orphan_requests_indices)

    if verbose > 1:
        for i in range(nranks):
            if rank == i:
                print(f"Debug Desc Exchange (to exchange), rank {i}")
                print(f" - send {orphan_requests_send_counts}")
                print(f" - recv {orphan_requests_recv_counts}")
                print(f"", flush=True)
            comm.Barrier()

    data_new = {}
    for k in data.keys():
        orphan_requests_send = data[k][orphan_requests_indices]
        orphan_requests_recv = np.empty(orphan_requests_recv_total, dtype=data[k].dtype)
        exchange_Alltoallv(
            [
                orphan_requests_send,
                (orphan_requests_send_counts, orphan_requests_send_offsets),
                orphan_requests_send.dtype.char,
            ],
            [
                orphan_requests_recv,
                (orphan_requests_recv_counts, orphan_requests_recv_offsets),
                orphan_requests_recv.dtype.char,
            ],
        )
        data_new[k] = np.concatenate(
            (data[k][~orphan_requests_mask], orphan_requests_recv)
        )

    if verbose > 1 and rank == 0:
        print("Exchange succeeded, verifying data integrity", flush=True)
    # comm.Barrier()

    # Verification
    localcount_after = len(data_new[key])
    localcount_missmatch = local_orphan_count[0]
    # calculate new missmatch
    my_data = np.unique(data_new[key])
    my_data = my_data[my_data >= 0]
    islocal = np.isin(my_data, local_keys, assume_unique=True)
    missing_keys = my_data[~islocal]
    localcount_missmatch_after = len(missing_keys)
    localcounts = np.array(
        [
            localcount,
            localcount_after,
            localcount_missmatch,
            localcount_missmatch_after,
        ],
        dtype=np.int64,
    )
    totalcounts = np.empty_like(localcounts)
    comm.Allreduce(localcounts, totalcounts)
    (
        totalcount_before,
        totalcount_after,
        totalcount_missmatch,
        totalcount_missmatch_after,
    ) = totalcounts

    if verbose and rank == 0:
        print(f"exchange summary ({'all2all' if do_all2all else 'neighbors'}):")
        print(
            f"   Ntot -> Ntot: {totalcount_before:10d} -> {totalcount_after:10d} "
            "(should remain the same)"
        )
        print(
            f"   Orph -> Orph: {totalcount_missmatch:10d} -> "
            f"{totalcount_missmatch_after:10d} (should be 0 after)"
        )
        print("", flush=True)

    # did we conserve number of particles?
    if rank == 0 and totalcount_before != totalcount_after:
        print(
            "Error in exchange: Lost halos during progenitor exchange: "
            f"{totalcount_before} -> {totalcount_after}",
            file=sys.stderr,
            flush=True,
        )
        comm.Abort()

    # if we were not able to assign all orphans to the neighbors, try all2all
    if not do_all2all and totalcount_missmatch_after > 0:
        if verbose and rank == 0:
            print(
                "exchange all2all since neighbor exchange was not able to assign all: "
                f"{totalcount_missmatch} -> {totalcount_missmatch_after}",
                flush=True,
            )
        return exchange(
            partition,
            data_new,
            key,
            local_keys,
            verbose=verbose,
            filter_key=filter_key,
            do_all2all=True,
            replace_notfound_key=replace_notfound_key,
        )

    # if we are still not able to assign all orphans, replace key or abort after
    # printing some debug messages
    if replace_notfound_key is not None and localcount_missmatch_after > 0:
        d = data_new[key]
        d[np.isin(d, missing_keys)] = replace_notfound_key
    for i in range(nranks):
        if rank == i and localcount_missmatch_after != 0 and verbose > 1:
            print(
                f"Warning from rank {rank} in exchange: Unable to assign all "
                f"progenitors to correct ranks (failed for "
                f"{localcount_missmatch_after} out of {localcount_missmatch})"
            )
            print("Could not assign keys: ", missing_keys)
            print("", flush=True)
        comm.Barrier()

    if rank == 0 and totalcount_missmatch_after != 0:
        if replace_notfound_key is None:
            print(
                f"Error in exchange: Unable to assign all progenitors to correct ranks "
                f"(tried to reassign {totalcount_missmatch}, failed for "
                f"{totalcount_missmatch_after})",
                file=sys.stderr,
                flush=True,
            )
            comm.Abort()
        elif verbose:
            print(
                f"Warning in exchange: Unable to assign all progenitors to correct "
                f"ranks (tried to reassign {totalcount_missmatch}, failed for "
                f"{totalcount_missmatch_after}), replacing missing values with "
                f"{replace_notfound_key}",
                flush=True,
            )

    return data_new



from typing import List, Mapping, Union

import numpy as np
import numba

from .s2_partition import S2Partition

ParticleDataT = Mapping[str, np.ndarray]


@numba.jit(nopython=True)
def _count_neighbors(
    theta: np.ndarray,
    phi: np.ndarray,
    overload_angle: float,
    my_rank: int,
    segment_theta_low: np.ndarray,
    segment_theta_high: np.ndarray,
    segment_phi_low: np.ndarray,
    segment_phi_high: np.ndarray,
    send_counts: np.ndarray,
    send_counts_by_rank: np.ndarray,
):
    npart = len(theta)
    nsegments = len(segment_theta_low)
    for i in range(npart):
        for j in range(nsegments):
            if j == my_rank:
                continue
            phi_i = phi[i]
            if phi_i < segment_phi_low[j] - overload_angle:
                phi_i += 2 * np.pi
            if phi_i >= segment_phi_high[j] + overload_angle:
                phi_i -= 2 * np.pi
            if (
                theta[i] >= segment_theta_low[j] - overload_angle
                and theta[i] < segment_theta_high[j] + overload_angle
                and phi_i >= segment_phi_low[j] - overload_angle
                and phi_i < segment_phi_high[j] + overload_angle
            ):
                send_counts[i] += 1
                send_counts_by_rank[j] += 1


@numba.jit(nopython=True)
def _calculate_partition(
    theta: np.ndarray,
    phi: np.ndarray,
    overload_angle: float,
    my_rank: int,
    segment_theta_low: np.ndarray,
    segment_theta_high: np.ndarray,
    segment_phi_low: np.ndarray,
    segment_phi_high: np.ndarray,
    send_offset_by_rank: np.ndarray,
    send_permutation: np.ndarray,
):
    send_count_by_rank = np.zeros_like(send_offset_by_rank)
    npart = len(theta)
    nsegments = len(segment_theta_low)
    for i in range(npart):
        for j in range(nsegments):
            if j == my_rank:
                continue
            phi_i = phi[i]
            if phi_i < segment_phi_low[j] - overload_angle:
                phi_i += 2 * np.pi
            if phi_i >= segment_phi_high[j] + overload_angle:
                phi_i -= 2 * np.pi
            if (
                theta[i] >= segment_theta_low[j] - overload_angle
                and theta[i] < segment_theta_high[j] + overload_angle
                and phi_i >= segment_phi_low[j] - overload_angle
                and phi_i < segment_phi_high[j] + overload_angle
            ):
                send_permutation[send_offset_by_rank[j] + send_count_by_rank[j]] = i
                send_count_by_rank[j] += 1


def s2_overload(
    partition: S2Partition,
    data: ParticleDataT,
    overload_angle: float,
    *,
    theta_key: str = "theta",
    phi_key: str = "phi",
    verbose: Union[bool, int] = False,
):
    """Copy data within an overload angle to the neighboring ranks ("ghost" particles)

    This method assumes that the particle data is already correctly distributed, i.e.
    that all particles on a given rank are within the bounds of the rank's segment.

    Parameters
    ----------
    partition:
        The MPI partition defining which rank should own which subvolume of the
        data

    data:
        The particle data to be redistributed, as a collection of 1-dimensional arrays.
        Each array must have the same length (number of particles) and the map needs
        to contain at least the keys `theta_key` and `phi_key`.

    overload_angle:
        The overload angle in radians. Particles within this angle of a rank's
        segment will be copied to the neighboring ranks. Note that the angle can be
        at maximum half of the smallest segment size in the partition.

    theta_key:
        The key in `data` that contains the particle theta coordinates (latitude),
        in the range [0, pi].

    phi_key:
        The key in `data` that contains the particle phi coordinates (longitude),
        in the range [0, 2*pi].

    verbose:
        If True, print summary statistics of the distribute. If > 1, print
        statistics of each rank (i.e. how much data each rank sends to every
        other rank).

    Returns
    -------
    data: ParticleDataT
        The combined data of objects within the rank's segmentt as well as the
        objects within the overload angle of neighboring ranks.

    Notes
    -----

    The function does not change the objects' coordinates or alter any data. Objects
    that have been overloaded accross the periodic boundary at 0 and 2pi will still have
    the original phi. In case "local" coordinates are required, this will need to be
    done manually after calling this function.
    """
    # verify data is normalized
    assert np.all(data[theta_key] >= 0)
    assert np.all(data[theta_key] <= np.pi)
    assert np.all(data[phi_key] >= 0)
    assert np.all(data[phi_key] <= 2 * np.pi)

    # count for each particle to many ranks it needs to be sent
    send_count_by_particle = np.zeros_like(data[theta_key], dtype=np.int32)
    send_counts = np.zeros(partition.nranks, dtype=np.int32)  # by rank

    segment_theta_low, segment_theta_high = partition.all_theta_extents.T
    segment_phi_low, segment_phi_high = partition.all_phi_extents.T

    _count_neighbors(
        data[theta_key],
        data[phi_key],
        overload_angle,
        partition.rank,
        segment_theta_low,
        segment_theta_high,
        segment_phi_low,
        segment_phi_high,
        send_count_by_particle,
        send_counts,
    )
    total_send_count = np.sum(send_count_by_particle)
    assert np.sum(send_counts) == total_send_count
    send_displacements = np.insert(np.cumsum(send_counts)[:-1], 0, 0)

    send_permutation = np.empty(total_send_count, dtype=np.int64)
    send_permutation[:] = -1

    _calculate_partition(
        data[theta_key],
        data[phi_key],
        overload_angle,
        partition.rank,
        segment_theta_low,
        segment_theta_high,
        segment_phi_low,
        segment_phi_high,
        send_displacements,
        send_permutation,
    )
    assert np.all(send_permutation >= 0)

    # Check how many elements will be received
    recv_counts = np.empty_like(send_counts)
    partition.comm.Alltoall(send_counts, recv_counts)
    recv_displacements = np.insert(np.cumsum(recv_counts)[:-1], 0, 0)
    total_receive_count = np.sum(recv_counts)

    # debug message
    if verbose > 1:
        for i in range(partition.nranks):
            if partition.rank == i:
                print(f"Overload Debug Rank {i}")
                print(f" - rank sends    {total_send_count:10d} particles")
                print(f" - rank receives {total_receive_count:10d} particles")
                print(f" - send_counts:        {send_counts}")
                print(f" - send_displacements: {send_displacements}")
                print(f" - recv_counts:        {recv_counts}")
                print(f" - recv_displacements: {recv_displacements}")
                print("", flush=True)
            partition.comm.Barrier()

    # send data all-to-all, each array individually
    data_new = {}

    for k in data.keys():
        # prepare send-array
        ds = data[k][send_permutation]
        # prepare recv-array
        dr = np.empty(total_receive_count, dtype=ds.dtype)
        # exchange data
        s_msg = [ds, (send_counts, send_displacements), ds.dtype.char]
        r_msg = [dr, (recv_counts, recv_displacements), ds.dtype.char]
        partition.comm.Alltoallv(s_msg, r_msg)
        # add received data to original data
        data_new[k] = np.concatenate((data[k], dr))

    return data_new



"""Top-level package for MPIPartition."""

__author__ = """Michael Buehlmann"""
__email__ = "buehlmann.michi@gmail.com"
__version__ = "1.1.3"


from .distribute import distribute
from .exchange import exchange
from .overload import overload
from .partition import Partition
from .spherical_partition import S2Partition, s2_distribute, s2_overload

__all__ = [
    "distribute",
    "exchange",
    "overload",
    "Partition",
    "S2Partition",
    "s2_distribute",
    "s2_overload",
]



import sys
from typing import List, Mapping, Union

import numpy as np

from .partition import MPI, Partition

ParticleDataT = Mapping[str, np.ndarray]


def distribute(
    partition: Partition,
    box_size: float,
    data: ParticleDataT,
    coord_keys: List[str],
    *,
    verbose: Union[bool, int] = False,
    verify_count: bool = True,
) -> ParticleDataT:
    """Distribute data among MPI ranks according to data position and volume partition

    The position of each TreeData element is given by the data columns
    specified with `coord_keys`.

    Parameters
    ----------

    partition:
        The MPI partition defining which rank should own which subvolume of the
        data

    box_size:
        The size of the full simulation volume

    data:
        The treenode / coretree data that should be distributed

    coord_keys:
        The columns in `data` that define the position of the object

    verbose:
        If True, print summary statistics of the distribute. If > 1, print
        statistics of each rank (i.e. how much data each rank sends to every
        other rank).

    verify_count:
        If True, make sure that total number of objects is conserved

    Returns
    -------
    data: ParticleDataT
        The distributed particle data (i.e. the data that this rank owns)

    """
    assert len(coord_keys) == partition.dimensions

    # get some MPI and partition parameters
    nranks = partition.nranks
    if nranks == 1:
        return data

    rank = partition.rank
    comm = partition.comm
    dimensions = partition.dimensions
    ranklist = np.array(partition.ranklist)
    extent = box_size * np.array(partition.extent)

    # count number of particles we have
    total_to_send = len(data[coord_keys[0]])

    if total_to_send > 0:
        # Check validity of coordinates
        for i in range(dimensions):
            _x = data[coord_keys[i]]
            _min = _x.min()
            _max = _x.max()
            if _min < 0 or _max > box_size:
                print(
                    f"Error in distribute: position {coord_keys[i]} out of range: [{_min}, {_max}]",
                    file=sys.stderr,
                    flush=True,
                )
                comm.Abort()

        # Find home of each particle
        idx = np.array(
            [data[coord_keys[i]] / extent[i] for i in range(dimensions)]
        ).astype(np.int32)
        idx = np.clip(idx, 0, np.array(partition.decomposition)[:, np.newaxis] - 1)
        home_idx = ranklist[tuple(idx)]
    else:
        # there are no particles on this rank
        home_idx = np.empty(0, dtype=np.int32)

    # sort by rank
    s = np.argsort(home_idx)
    home_idx = home_idx[s]

    # offsets and counts
    send_displacements = np.searchsorted(home_idx, np.arange(nranks))
    send_displacements = send_displacements.astype(np.int32)
    send_counts = np.append(send_displacements[1:], total_to_send) - send_displacements
    send_counts = send_counts.astype(np.int32)

    # announce to each rank how many objects will be sent
    recv_counts = np.empty_like(send_counts)
    comm.Alltoall(send_counts, recv_counts)
    recv_displacements = np.insert(np.cumsum(recv_counts)[:-1], 0, 0)

    # number of objects that this rank will receive
    total_to_receive = np.sum(recv_counts)

    # debug message
    if verbose > 1:
        for i in range(nranks):
            if rank == i:
                print(f"Distribute Debug Rank {i}")
                print(f" - rank has {total_to_send} particles")
                print(f" - rank receives {total_to_receive} particles")
                print(f" - send_counts:        {send_counts}")
                print(f" - send_displacements: {send_displacements}")
                print(f" - recv_counts:        {recv_counts}")
                print(f" - recv_displacements: {recv_displacements}")
                print(f"", flush=True)
            comm.Barrier()

    # send data all-to-all, each array individually
    data_new = {k: np.empty(total_to_receive, dtype=data[k].dtype) for k in data.keys()}

    for k in data.keys():
        d = data[k][s]
        s_msg = [d, (send_counts, send_displacements), d.dtype.char]
        r_msg = [data_new[k], (recv_counts, recv_displacements), d.dtype.char]
        comm.Alltoallv(s_msg, r_msg)

    if verify_count:
        local_counts = np.array(
            [len(data[coord_keys[0]]), len(data_new[coord_keys[0]])], dtype=np.int64
        )
        global_counts = np.empty_like(local_counts)
        comm.Reduce(local_counts, global_counts, op=MPI.SUM, root=0)
        if rank == 0 and global_counts[0] != global_counts[1]:
            print(
                f"Error in distribute: particle count during distribute was not maintained ({global_counts[0]} -> {global_counts[1]})",
                file=sys.stderr,
                flush=True,
            )
            comm.Abort()

    return data_new



import numpy as np
import numpy.typing as npt
from typing import Optional, Tuple, List
from dataclasses import dataclass

from mpi4py import MPI


def _cap_area(theta):
    h = 1 - np.cos(theta)
    return 2 * np.pi * h


def _cap_angle(area):
    h = area / (2 * np.pi)
    return np.arccos(1 - h)


def _s2_partition(n, adjust_theta=True):
    # Following the algorithm by Paul Leopardi (https://eqsp.sourceforge.net/)
    # TODO: generalize to arbitrary number of dimensions
    assert n > 1
    if n == 2:
        return (
            np.pi / 2,
            np.array([np.pi / 2], dtype=np.float64),
            np.array([], dtype=np.int32),
        )

    target_area = 4 * np.pi / n

    # 1: polar caps
    theta_c = _cap_angle(target_area)

    # 2-4: rings
    delta_i = np.sqrt(target_area)  # ideal spacing
    n_rings_i = (np.pi - 2 * theta_c) / delta_i
    n_rings = max(int(n_rings_i + 0.5), 1)
    # 5-6: ring segments
    delta_f = (np.pi - 2 * theta_c) / n_rings
    theta_f = theta_c + np.arange(n_rings) * delta_f
    theta_f = np.append(theta_f, np.pi - theta_c)

    n_segments_i = (_cap_area(theta_f[1:]) - _cap_area(theta_f[:-1])) / target_area
    n_segments = np.zeros(n_rings, dtype=np.int32)
    remainder = 0
    for i in range(n_rings):
        ni = int(n_segments_i[i] + remainder + 0.5)
        remainder += n_segments_i[i] - ni
        n_segments[i] = ni

    assert abs(remainder) < 1e-6
    assert np.sum(n_segments) + 2 == n

    # 7: adjust theta_f for equal area
    if adjust_theta:
        areas = target_area * np.array(n_segments)
        cum_areas = target_area + np.cumsum(areas)
        theta_f[1:] = _cap_angle(cum_areas)

    return theta_c, theta_f, n_segments


@dataclass
class S2Segment:
    theta_range: Tuple[float]
    phi_range: Tuple[float]
    area: float
    edge_length: float


def _build_s2_segment_list(theta_cap, ring_thetas, ring_segments):
    segments = []

    # cap
    cap_area = _cap_area(theta_cap)
    cap_edge_length = 2 * np.pi * np.sin(theta_cap)
    segments.append(
        S2Segment((0, theta_cap), (0, 2 * np.pi), cap_area, cap_edge_length)
    )

    # rings
    for i in range(len(ring_segments)):
        theta_start = ring_thetas[i]
        theta_end = ring_thetas[i + 1]
        area = (_cap_area(theta_end) - _cap_area(theta_start)) / ring_segments[i]
        phi_edges = np.linspace(0, 2 * np.pi, ring_segments[i] + 1, endpoint=True)
        edge_length_top = 2 * np.pi * np.sin(theta_start) / ring_segments[i]
        edge_length_bottom = 2 * np.pi * np.sin(theta_end) / ring_segments[i]
        edge_length = (
            edge_length_top + edge_length_bottom + 2 * (theta_end - theta_start)
        )
        for j in range(ring_segments[i]):
            segments.append(
                S2Segment(
                    (theta_start, theta_end),
                    (phi_edges[j], phi_edges[j + 1]),
                    area,
                    edge_length,
                )
            )

    # cap
    segments.append(
        S2Segment((np.pi - theta_cap, np.pi), (0, 2 * np.pi), cap_area, cap_edge_length)
    )

    return segments


def _print_segmentation_info(
    nranks, theta_cap, ring_thetas, ring_segments, precision=3
):
    print(f"Segmentation statistics for {nranks} ranks:")
    print(f"  polar cap angle: {theta_cap:.{precision}f}")
    print(f"  number of rings: {len(ring_segments)}")
    for i in range(len(ring_segments)):
        print(
            f"    ring {i:3d}: {ring_segments[i]:3d} segments between "
            f"theta=[{ring_thetas[i]:.{precision}f}, {ring_thetas[i+1]:.{precision}f}]]"
        )


# area imbalance
def _print_area_imabalance(segments: List[S2Segment], precision=3):
    areas = np.array([r.area for r in segments])
    assert np.isclose(np.sum(areas), 4 * np.pi)
    print("  Segment area imbalance:")
    print(f"    max/min: {np.max(areas) / np.min(areas):.{precision}f}")
    print(f"    max/avg: {np.max(areas) / np.mean(areas):.{precision}f}")


def _print_edge_to_area_ratio(segments: List[S2Segment], precision=3):
    areas = np.array([r.area for r in segments])
    edge_lengths = np.array([r.edge_length for r in segments])
    total_ratio = np.sum(edge_lengths) / np.sum(areas)
    print(f"  Total edge/area ratio: {total_ratio:.{precision}f}")


class S2Partition:
    """An MPI decomposition of the spherical shell into equal-area segments

    Parameters
    ----------
    equal_area : bool
        If True, the spherical shell is divided into equal-area segments. If False, use
        equally spaced rings (in theta) instead.

    comm : MPI.Comm
        The MPI communicator to use for the decomposition (default: COMM_WORLD)

    verbose : bool
        If True, rank 0 will print information about the segmentation.

    """

    # parition properties
    _comm: MPI.Comm
    _nranks: int
    _theta_cap: float
    _ring_thetas: npt.NDArray[np.float64]
    _ring_segments: npt.NDArray[np.int64]
    _equal_area: bool
    _ring_dtheta: Optional[float]
    _all_s2_segments: List[S2Segment]

    # rank properties
    _rank: int
    _s2_segment: S2Segment

    @property
    def comm(self):
        """MPI Communicator"""
        return self._comm

    @property
    def rank(self) -> int:
        """the MPI rank of this processor"""
        return self._rank

    @property
    def nranks(self) -> int:
        """the total number of processors"""
        return self._nranks

    @property
    def equal_area(self) -> bool:
        """whether the partition is equal-area"""
        return self._equal_area

    @property
    def theta_cap(self) -> float:
        """the polar cap angle"""
        return self._theta_cap

    @property
    def ring_thetas(self) -> npt.NDArray[np.float64]:
        """the theta boundaries of all rings"""
        return self._ring_thetas

    @property
    def ring_segments(self) -> npt.NDArray[np.int64]:
        """the number of segments in each ring"""
        return self._ring_segments

    @property
    def ring_dtheta(self) -> Optional[float]:
        """the theta spacing between rings (only for non-equal-area partitions)"""
        return self._ring_dtheta

    @property
    def phi_extent(self) -> Tuple[float, float]:
        """the phi extent of the segment assigned to this rank"""
        return self._s2_segment.phi_range

    @property
    def theta_extent(self) -> Tuple[float, float]:
        """the theta extent of the segment assigned to this rank"""
        return self._s2_segment.theta_range

    @property
    def area(self) -> float:
        """the area of the segment assigned to this rank"""
        return self._s2_segment.area

    @property
    def all_phi_extents(self) -> npt.NDArray[np.float64]:
        """the phi extent of all segments, shape (nranks, 2)"""
        return np.array([s.phi_range for s in self._all_s2_segments])

    @property
    def all_theta_extents(self) -> npt.NDArray[np.float64]:
        """the theta extent of all segments, shape (nranks, 2)"""
        return np.array([s.theta_range for s in self._all_s2_segments])

    def __init__(
        self,
        *,
        equal_area: bool = True,
        comm: MPI.Comm = MPI.COMM_WORLD,
        verbose: bool = False,
    ):
        self._comm = comm
        self._rank = self._comm.Get_rank()
        self._nranks = self._comm.Get_size()
        self._equal_area = equal_area
        self._theta_cap, self._ring_thetas, self._ring_segments = _s2_partition(
            self._nranks, equal_area
        )
        if not equal_area:
            if len(self._ring_thetas) > 1:
                self._ring_dtheta = self._ring_thetas[1] - self._ring_thetas[0]
            else:
                self._ring_dtheta = 0.0
        else:
            self._ring_dtheta = None

        self._all_s2_segments = _build_s2_segment_list(
            self._theta_cap, self._ring_thetas, self._ring_segments
        )
        assert len(self._all_s2_segments) == self._nranks
        self._s2_segment = self._all_s2_segments[self._rank]

        if verbose and self._rank == 0:
            _print_segmentation_info(
                self._nranks,
                self._theta_cap,
                self._ring_thetas,
                self._ring_segments,
            )
            _print_area_imabalance(self._all_s2_segments)
            _print_edge_to_area_ratio(self._all_s2_segments)
            print()


def visualize_s2_partition(
    nranks: int, equal_area: bool = True, use_mollweide: bool = True, fig=None
):
    """Visualize the S2 partitioning of the sphere.

    Parameters
    ----------
    nranks : int
        Number of ranks to partition the sphere into.
    equal_area : bool, optional
        If True, partition the sphere into equal area regions by adjusting theta.
        Otherwise, keep delta_theta of rings constant.
    use_mollweide : bool, optional
        If True, use the Mollweide projection. Otherwise, use a regular plot.
    fig : matplotlib.figure.Figure, optional
        Figure to plot on. If None, create a new figure.

    Returns
    -------
    matplotlib.figure.Figure
        Figure containing the plot.
    matplotlib.axes.Axes
        Axes containing the plot.
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("matplotlib is required for visualization")

    if fig is None:
        fig = plt.figure()
    if use_mollweide:
        ax = fig.add_subplot(1, 1, 1, projection="mollweide")
    else:
        ax = fig.add_subplot(1, 1, 1)

    theta_c, theta_f, n_segments = _s2_partition(nranks, equal_area)

    for i, nsegments in enumerate(n_segments):
        theta_lo = theta_f[i]
        theta_hi = theta_f[i + 1]
        if use_mollweide:
            theta_lo -= np.pi / 2
            theta_hi -= np.pi / 2

        # draw vertical bars
        for j in range(nsegments):
            ax.plot(
                (2 * np.pi / nsegments * j - np.pi) * np.ones(2),
                [theta_lo, theta_hi],
                color="black",
                linewidth=0.5,
            )
        # draw upper ring
        ax.axhline(theta_lo, color="black", linewidth=0.5)
    # draw final ring (only if we have rings!)
    if n_segments.size:
        ax.axhline(theta_hi, color="black", linewidth=0.5)
    else:
        ax.axhline(theta_c, color="black", linewidth=0.5)

    if not use_mollweide:
        ax.set(
            xlabel=r"$\phi$",
            ylabel=r"$\theta$",
            xlim=(-np.pi, np.pi),
            ylim=(0, np.pi),
        )

    return fig, ax



import sys
from typing import Optional

import click

from mpipartition.spherical_partition.s2_partition import (
    _s2_partition,
    visualize_s2_partition,
    _build_s2_segment_list,
    _print_area_imabalance,
    _print_edge_to_area_ratio,
    _print_segmentation_info,
)


@click.command()
@click.argument(
    "nranks",
    type=int,
)
@click.option(
    "--equal-area",
    is_flag=True,
    help=(
        "If set, partition S2 into equal area segments by varying delta_theta of "
        "rings."
    ),
)
@click.option(
    "--precision",
    type=int,
    default=6,
    help="Number of decimal places to use for printing.",
)
@click.option(
    "--figure",
    type=click.Path(dir_okay=False, writable=True),
    help="If set, save a visual representation of the S2 partitioning to this file.",
)
@click.option(
    "--use-mollweide",
    is_flag=True,
    help=(
        "If set, use the Mollweide projection. Otherwise, use a rectangular theta-phi "
        "plot for visualization."
    ),
)
@click.option(
    "--figure-pad",
    type=float,
    help=(
        "pad_inches argument used for saving the figure. Set to 0 to remove whitespace."
    ),
)
def cli(
    nranks: int,
    equal_area: bool,
    precision: int,
    figure: Optional[str],
    use_mollweide: bool,
    figure_pad: float,
):
    theta_cap, ring_thetas, ring_segments = _s2_partition(nranks, equal_area)
    all_s2_segments = _build_s2_segment_list(theta_cap, ring_thetas, ring_segments)
    _print_segmentation_info(
        nranks,
        theta_cap,
        ring_thetas,
        ring_segments,
        precision=precision,
    )
    _print_area_imabalance(all_s2_segments, precision=precision)
    _print_edge_to_area_ratio(all_s2_segments, precision=precision)

    if figure is not None:
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            print(
                "matplotlib is required for visualization, install mpipartition with "
                "`pip install mpipartition[viz]",
                file=sys.stderr,
            )
            return
        fig, ax = visualize_s2_partition(nranks, equal_area, use_mollweide)
        if use_mollweide:
            ax.set(xticklabels=[], yticklabels=[])
        fig.savefig(figure, bbox_inches="tight", pad_inches=figure_pad)



from typing import Mapping, List, Union
import numpy as np
import sys

from .s2_partition import S2Partition
from mpi4py import MPI

ParticleDataT = Mapping[str, np.ndarray]


def s2_distribute(
    partition: S2Partition,
    data: ParticleDataT,
    *,
    theta_key: str = "theta",
    phi_key: str = "phi",
    verbose: Union[bool, int] = False,
    verify_count: bool = True,
    validate_home: bool = False,
) -> ParticleDataT:
    """Distribute particles among MPI ranks according to the S2 partition.

    Parameters
    ----------
    partition:
        The S2 partition to use for the distribution.

    data:
        The particle data to distribute, as a collection of 1-dimensional arrays.
        Each array must have the same length (number of particles) and the map needs
        to contain at least the keys `theta_key` and `phi_key`.

    theta_key:
        The key in `data` that contains the particle theta coordinates (latitude),
        in the range [0, pi].

    phi_key:
        The key in `data` that contains the particle phi coordinates (longitude),
        in the range [0, 2*pi].

    verbose:
        If True, print summary statistics of the distribute. If > 1, print
        statistics of each rank (i.e. how much data each rank sends to every
        other rank).

    verify_count:
        If True, make sure that total number of objects is conserved.

    validate_home:
        If True, validate that each rank indeed owns the particles that it was sent.

    Returns
    -------
    data: ParticleDataT
        The distributed particle data (i.e. the data that this rank owns)

    """
    # count number of particles we have
    total_to_send = len(data[theta_key])

    # verify data is normalized
    assert np.all(data[theta_key] >= 0)
    assert np.all(data[theta_key] < np.pi)
    assert np.all(data[phi_key] >= 0)
    assert np.all(data[phi_key] < 2 * np.pi)

    # ring idx: 0=cap, 1=first ring, 2=second ring, etc.
    if partition.equal_area:
        ring_idx = np.digitize(data[theta_key], partition.ring_thetas)
    else:  # equal theta
        if partition.nranks == 2:
            ring_idx = (data[theta_key] > partition.theta_cap).astype(np.int32)
        else:
            assert partition.ring_dtheta is not None
            ring_idx = (
                (data[theta_key] - partition.theta_cap) // partition.ring_dtheta
            ).astype(np.int32) + 1
            ring_idx = np.clip(ring_idx, 0, len(partition.ring_segments) + 1)

    phi_idx = np.zeros_like(ring_idx, dtype=np.int32)
    mask_is_on_ring = (ring_idx > 0) & (ring_idx <= len(partition.ring_segments))
    phi_idx[mask_is_on_ring] = (
        data[phi_key][mask_is_on_ring]
        / (2 * np.pi)
        * partition.ring_segments[ring_idx[mask_is_on_ring] - 1]
    ).astype(np.int32)

    # rank index where each ring starts
    ring_start_idx = np.zeros(len(partition.ring_segments) + 2, dtype=np.int32)
    ring_start_idx[1] = 1
    ring_start_idx[2:] = np.cumsum(partition.ring_segments) + 1

    # rank index of each particle
    home_idx = ring_start_idx[ring_idx] + phi_idx

    assert np.all(home_idx >= 0)
    assert np.all(home_idx < partition.nranks)

    # TODO: this code is duplicated in distribute.py – unify!

    # sort by rank
    s = np.argsort(home_idx)
    home_idx = home_idx[s]

    # offsets and counts
    send_displacements = np.searchsorted(home_idx, np.arange(partition.nranks))
    send_displacements = send_displacements.astype(np.int32)
    send_counts = np.append(send_displacements[1:], total_to_send) - send_displacements
    send_counts = send_counts.astype(np.int32)

    # announce to each rank how many objects will be sent
    recv_counts = np.empty_like(send_counts)
    partition.comm.Alltoall(send_counts, recv_counts)
    recv_displacements = np.insert(np.cumsum(recv_counts)[:-1], 0, 0)

    # number of objects that this rank will receive
    total_to_receive = np.sum(recv_counts)

    # debug message
    if verbose > 1:
        for i in range(partition.nranks):
            if partition.rank == i:
                print(f"Distribute Debug Rank {i}")
                print(f" - rank has {total_to_send} particles")
                print(f" - rank receives {total_to_receive} particles")
                print(f" - send_counts:        {send_counts}")
                print(f" - send_displacements: {send_displacements}")
                print(f" - recv_counts:        {recv_counts}")
                print(f" - recv_displacements: {recv_displacements}")
                print(f"", flush=True)
            partition.comm.Barrier()

    # send data all-to-all, each array individually
    data_new = {k: np.empty(total_to_receive, dtype=data[k].dtype) for k in data.keys()}

    for k in data.keys():
        d = data[k][s]
        s_msg = [d, (send_counts, send_displacements), d.dtype.char]
        r_msg = [data_new[k], (recv_counts, recv_displacements), d.dtype.char]
        partition.comm.Alltoallv(s_msg, r_msg)

    if verify_count:
        local_counts = np.array(
            [len(data[theta_key]), len(data_new[theta_key])], dtype=np.int64
        )
        global_counts = np.empty_like(local_counts)
        partition.comm.Reduce(local_counts, global_counts, op=MPI.SUM, root=0)
        if partition.rank == 0 and global_counts[0] != global_counts[1]:
            print(
                f"Error in distribute: particle count during distribute was not "
                f"maintained ({global_counts[0]} -> {global_counts[1]})",
                file=sys.stderr,
                flush=True,
            )
            partition.comm.Abort()

    if validate_home:
        assert np.all(data_new[theta_key] >= partition.theta_extent[0])
        assert np.all(data_new[theta_key] < partition.theta_extent[1])
        assert np.all(data_new[phi_key] >= partition.phi_extent[0])
        assert np.all(data_new[phi_key] < partition.phi_extent[1])

    return data_new



import itertools
from typing import List, Mapping, Union

import numpy as np

from .partition import Partition

ParticleDataT = Mapping[str, np.ndarray]


def overload(
    partition: Partition,
    box_size: float,
    data: ParticleDataT,
    overload_length: float,
    coord_keys: List[str],
    *,
    verbose: Union[bool, int] = False,
):
    """Copy data within an overload length to the neighboring ranks

    This method assumes that the volume cube is periodic and will wrap the data
    around the boundary interfaces.

    Parameters
    ----------
    partition:
        The MPI partition defining which rank should own which subvolume of the
        data

    box_size:
        the size of the full volume cube

    data:
        The treenode / coretree data that should be distributed

    overload_length:
        The thickness of the boundary layer that will be copied to the
        neighboring rank. Must be smaller than half the extent of the local
        subvolume (along any axis)

    coord_keys:
        The columns in `data` that define the position of the object

    verbose:
        If True, print summary statistics of the distribute. If > 1, print
        statistics of each rank (i.e. how much data each rank sends to every
        other rank).

    Returns
    -------
    data: TreeDataT
        The combined data of objects within the rank's subvolume as well as the
        objects within the overload region of neighboring ranks

    Notes
    -----

    The function does not change the objects' coordinates or alter any data.
    Objects that have been overloaded accross the periodic boundaries will still
    have the original positions. In case "local" coordinates are required, this
    will need to be done manually after calling this function.

    """
    assert len(coord_keys) == partition.dimensions
    for i in range(partition.dimensions):
        # currently can't overload if only 1 rank
        assert partition.decomposition[i] > 1
        # can maximally overload half the local extent, otherwise would need
        # to send same particle multiple times
        max_overload = 0.5 * partition.extent[i] * box_size
        assert overload_length < max_overload

    nranks = partition.nranks
    if nranks == 1:
        return data

    rank = partition.rank
    comm = partition.comm
    dimensions = partition.dimensions
    origin = box_size * np.array(partition.origin)
    extent = box_size * np.array(partition.extent)

    neighbors = partition.neighbors

    # Find all overload regions each particle should be in
    overload = {}
    for i, x in enumerate(coord_keys):
        _i = np.zeros_like(data[x], dtype=np.int8)
        _i[data[x] < origin[i] + overload_length] = -1
        _i[data[x] > origin[i] + extent[i] - overload_length] = 1
        overload[i] = _i

    # Get particle indices of each of the 27 neighbors overload
    exchange_indices = [np.empty(0, dtype=np.int64) for i in range(nranks)]

    def add_exchange_indices(mask, idx):
        assert len(idx) == dimensions
        n = neighbors[tuple(_i + 1 for _i in idx)]
        if n != rank:
            exchange_indices[n] = np.union1d(exchange_indices[n], np.nonzero(mask)[0])

    corners = itertools.product([0, -1, 1], repeat=partition.dimensions)
    # skip first: will be [0,0,0]
    next(corners)

    for corner in corners:
        mask = np.ones_like(overload[0], dtype=np.bool_)
        for d in range(partition.dimensions):
            if corner[d] == 0:
                continue
            mask &= overload[d] == corner[d]
        add_exchange_indices(mask, corner)

    # Check how many elements will be sent
    send_counts = np.array([len(i) for i in exchange_indices], dtype=np.int32)
    send_idx = np.concatenate(exchange_indices)
    send_displacements = np.insert(np.cumsum(send_counts)[:-1], 0, 0)
    total_to_send = np.sum(send_counts)

    # Check how many elements will be received
    recv_counts = np.empty_like(send_counts)
    comm.Alltoall(send_counts, recv_counts)
    recv_displacements = np.insert(np.cumsum(recv_counts)[:-1], 0, 0)
    total_to_receive = np.sum(recv_counts)

    # debug message
    if verbose > 1:
        for i in range(nranks):
            if rank == i:
                print(f"Overload Debug Rank {i}")
                print(f" - rank sends    {total_to_send:10d} particles")
                print(f" - rank receives {total_to_receive:10d} particles")
                print(f" - send_counts:        {send_counts}")
                print(f" - send_displacements: {send_displacements}")
                print(f" - recv_counts:        {recv_counts}")
                print(f" - recv_displacements: {recv_displacements}")
                for i, x in enumerate(coord_keys):
                    print(f" - overload_{x}: {overload[i]}")
                print(f" - send_idx: {send_idx}")
                print(f"", flush=True)
            comm.Barrier()

    # send data all-to-all, each array individually
    data_new = {}

    for k in data.keys():
        # prepare send-array
        ds = data[k][send_idx]
        # prepare recv-array
        dr = np.empty(total_to_receive, dtype=ds.dtype)
        # exchange data
        s_msg = [ds, (send_counts, send_displacements), ds.dtype.char]
        r_msg = [dr, (recv_counts, recv_displacements), ds.dtype.char]
        comm.Alltoallv(s_msg, r_msg)
        # add received data to original data
        data_new[k] = np.concatenate((data[k], dr))

    return data_new



"""MPI Partitioning of a cube

"""

import itertools
import sys
from typing import List

import numpy as np
from mpi4py import MPI

_comm = MPI.COMM_WORLD
_rank = _comm.Get_rank()
_nranks = _comm.Get_size()


def _factorize(n):
    i = 2
    factors = []
    while i <= n:
        if (n % i) == 0:
            factors.append(i)
            n /= i
        else:
            i = i + 1
    return factors


def _distribute_factors(factors, target):
    current_topo = np.ones_like(target)
    remaining_topo = np.copy(np.array(target))
    for f in factors[::-1]:
        commensurate = (remaining_topo % f) == 0
        if not np.any(commensurate):
            raise RuntimeError(
                "commensurate topology impossible with given rank number and target topo"
            )
        # add to lowest possible number
        s = np.argsort(current_topo)
        idx = s[np.nonzero(commensurate[s])[0][0]]
        current_topo[idx] *= f
        remaining_topo[idx] /= f
    return current_topo, remaining_topo


class Partition:
    """An MPI partition of a cubic volume

    Parameters
    ----------

    dimension: int
        Numer of dimensions of the volume cube. Default: 3

    create_neighbor_topo : boolean
        If `True`, an additional graph communicator will be initialized
        connecting all direct neighbors (3**dimension - 1) symmetrically

    commensurate_topo : List[int]
        A proportional target topology for decomposition. When specified, a partition
        will be created so that `commensurate_topo[i] % partition.decomposition[i] == 0`
        for all `i`. The code will raise a RuntimeError if such a decomposition is not
        possible.

    Examples
    --------

    Using Partition on 8 MPI ranks to split a periodic unit-cube

    >>> partition = Partition(1.0)
    >>> partition.rank
    0
    >>> partition.decomposition
    np.ndarray([2, 2, 2])
    >>> partition.coordinates
    np.ndarray([0, 0, 0])
    >>> partition.origin
    np.ndarray([0., 0., 0.])
    >>> partition.extent
    np.ndarray([0.5, 0.5, 0.5])


    """

    def __init__(
        self,
        dimensions=3,
        *,
        create_neighbor_topo: bool = False,
        commensurate_topo: List[int] = None,
    ):
        self._topo = None
        self._neighbor_topo = None
        self._neighbor_ranks = None

        self._dimensions = dimensions
        self._rank = _rank
        self._nranks = _nranks

        assert dimensions > 0
        assert type(dimensions) == int

        if commensurate_topo is None:
            self._decomposition = MPI.Compute_dims(_nranks, [0] * self._dimensions)
        else:
            nranks_factors = _factorize(self._nranks)
            decomposition, remainder = _distribute_factors(
                nranks_factors, commensurate_topo
            )
            assert np.all(decomposition * remainder == np.array(commensurate_topo))
            assert np.prod(decomposition) == self._nranks
            self._decomposition = decomposition.tolist()

        periodic = [True] * self._dimensions

        self._topo = _comm.Create_cart(self._decomposition, periods=periodic)
        self._coords = list(self._topo.coords)

        self._neighbors = np.zeros([3] * self._dimensions, dtype=np.int32)
        for idx in itertools.product([-1, 0, 1], repeat=self._dimensions):
            coord = [
                (self._coords[d] + idx[d]) % self._decomposition[d]
                for d in range(self._dimensions)
            ]
            neigh = self._topo.Get_cart_rank(coord)
            self._neighbors[tuple(_i + 1 for _i in idx)] = neigh

        self._extent = [1.0 / self._decomposition[i] for i in range(self._dimensions)]
        self._origin = [
            self._coords[i] * self._extent[i] for i in range(self._dimensions)
        ]

        # A graph topology linking all neighbors
        if create_neighbor_topo:
            neighbors = np.unique(
                [n for n in self._neighbors.flatten() if n != self._rank]
            ).astype(np.int32)
            self._neighbor_topo = self._topo.Create_dist_graph_adjacent(
                sources=neighbors, destinations=neighbors, reorder=False
            )
            assert self._neighbor_topo.is_topo
            inout_neighbors = self._neighbor_topo.inoutedges
            assert len(inout_neighbors[0]) == len(inout_neighbors[1])
            for i in range(len(inout_neighbors[0])):
                if inout_neighbors[0][i] != inout_neighbors[1][i]:
                    print(
                        "neighbor topo: neighbors in sources and destinations are not ordered the same",
                        file=sys.stderr,
                        flush=True,
                    )
                    self._topo.Abort()
            self._neighbor_ranks = inout_neighbors[0]

    def __del__(self):
        if self._neighbor_topo is not None:
            self._neighbor_topo.Free()
        if self._topo is not None:
            self._topo.Free()

    @property
    def dimensions(self):
        """Dimension of the partitioned volume"""
        return self._dimensions

    @property
    def comm(self):
        """3D Cartesian MPI Topology / Communicator"""
        return self._topo

    @property
    def comm_neighbor(self):
        """Graph MPI Topology / Communicator, connecting the neighboring ranks
        (symmetric)"""
        return self._neighbor_topo

    @property
    def rank(self):
        """int: the MPI rank of this processor"""
        return self._topo.rank

    @property
    def nranks(self):
        """int: the total number of processors"""
        return self._nranks

    @property
    def decomposition(self):
        """np.ndarray: the decomposition of the cubic volume: number of ranks along each dimension"""
        return self._decomposition

    @property
    def coordinates(self):
        """np.ndarray: 3D indices of this processor"""
        return self._coords

    @property
    def extent(self):
        """np.ndarray: Length along each axis of this processors subvolume (same for all procs)"""
        return self._extent

    @property
    def origin(self) -> np.ndarray:
        """np.ndarray: Cartesian coordinates of the origin of this processor"""
        return self._origin

    def get_neighbor(self, di: List[int]) -> int:
        """get the rank of the neighbor at relative position (dx, dy, dz, ...)

        Parameters
        ----------

        di: List[int]
            list of relative coordinates, one of `[-1, 0, 1]`.
        """
        assert len(di) == self._dimensions
        return self._neighbors[np.array(di) + 1]

    @property
    def neighbors(self):
        """np.ndarray: a 3^d dimensional array with the ranks of the neighboring processes
        (`neighbors[1,1,1, ...]` is this processor)"""
        return self._neighbors

    @property
    def neighbor_ranks(self):
        """np.ndarray: a flattened list of the unique neighboring ranks"""
        return self._neighbor_ranks

    @property
    def ranklist(self):
        """np.ndarray: A complete list of ranks, aranged by their coordinates.
        The array has shape `partition.decomposition`"""
        ranklist = np.empty(self.decomposition, dtype=np.int32)
        for idx in itertools.product(*map(range, self.decomposition)):
            ranklist[tuple(idx)] = self._topo.Get_cart_rank(idx)
        return ranklist
